-- this is a fix for databases that were migrated by 2.7 but then accessed again with 2.6 which recreated it
-- the expenses table is not supposed to exist
DROP TABLE expense IF EXISTS;

-- this is a fix that makes sure items use a catalog item of the same type, in other words, service items use only catalog service items and so on
-- it is possible that a service item is assigned a catalog item of a different type like product catalog item or expense catalog item
UPDATE item SET item.catalogitemid=null WHERE item.catalogitemid IS NOT NULL AND NOT EXISTS (SELECT itemtype from catalogitem WHERE item.itemtype = catalogitem.itemtype AND item.catalogitemid = catalogitem.catalogitemid);

-- create new expense and expensecategory tables
CREATE TABLE expensecategory(
	expensecategoryid BIGINT GENERATED BY DEFAULT AS IDENTITY(START WITH 1) NOT NULL,
	uuid VARCHAR(36) NOT NULL,
	created TIMESTAMP NOT NULL,
	updated TIMESTAMP,
	businessid BIGINT,
	name VARCHAR(255) NOT NULL,
	enabled BOOLEAN,
	PRIMARY KEY(expensecategoryid),
	CONSTRAINT fk_expensecategory_businessid FOREIGN KEY (businessid) REFERENCES PUBLIC.business(businessid)
);

CREATE TABLE expense (
	expenseid BIGINT GENERATED BY DEFAULT AS IDENTITY(START WITH 1) NOT NULL,
	uuid VARCHAR(36) NOT NULL,
	created TIMESTAMP NOT NULL,
	updated TIMESTAMP,
	projectid BIGINT,
	date TIMESTAMP,
	expensecategoryid BIGINT,
	amount NUMERIC,
	currency VARCHAR(255),
	description VARCHAR(2000),
	notes VARCHAR(2000),
	tags VARCHAR(255),
	reference VARCHAR(100),
	billable BOOLEAN,
	taxexempt BOOLEAN,
	billabletotalamount NUMERIC,
	billabletotalcurrency VARCHAR(255),
	itemid BIGINT,
	PRIMARY KEY(expenseid),
	CONSTRAINT fk_expense_projectid FOREIGN KEY (projectid) REFERENCES PUBLIC.project(projectid),
	CONSTRAINT fk_expense_expensecategoryid FOREIGN KEY (expensecategoryid) REFERENCES PUBLIC.expensecategory(expensecategoryid),
	CONSTRAINT fk_expense_itemid FOREIGN KEY (itemid) REFERENCES PUBLIC.item(itemid)
);

-- link expenses to items
ALTER TABLE item ADD expenseid BIGINT;
ALTER TABLE item ADD CONSTRAINT fk_item_expenseid FOREIGN KEY (expenseid) REFERENCES PUBLIC.expense(expenseid);

-- create expense categories based on expense catalog items
-- at this point service = 0; expense = 1; product = 2
INSERT INTO expensecategory (expensecategoryid,uuid,created,updated,businessid,name,enabled) 
SELECT catalogitemid,uuid,created,updated,businessid,name,enabled
FROM catalogitem
WHERE catalogitem.itemtype=1;

-- create expense objects based on expense items
-- item.name + " " + item.description => expense.description
INSERT INTO expense (expenseid,uuid,created,updated,projectid,date,expensecategoryid,amount,currency,description,notes,billable,taxexempt,billabletotalamount,billabletotalcurrency) 
SELECT itemid,uuid,created,updated,projectid,date,catalogitemid,units*cost,costcurrency,name || ' ' || description,notes,billable,taxexempt,units*rate,costcurrency
FROM item
WHERE item.itemtype=1;

UPDATE expense SET itemid=(SELECT item.itemid FROM item WHERE expense.expenseid=item.itemid AND item.invoiceid is not null AND item.itemtype=1);
UPDATE item SET expenseid=itemid WHERE invoiceid is not null AND itemtype=1;
UPDATE item SET projectid=null WHERE invoiceid is not null AND itemtype=1;

-- delete not invoiced expense items because they were replaced by expenses
DELETE FROM item WHERE invoiceid is null AND itemtype=1;

-- set catalog item to null for expense items because expense items will no longer use catalog items
UPDATE item SET catalogitemid=null WHERE itemtype=1;

-- delete expense catalog items because they were replaced by expense categories
DELETE FROM catalogitem WHERE itemtype=1;

-- remove project billing setting used for expense items because they are not used anymore
ALTER TABLE project DROP COLUMN expensecost;
ALTER TABLE project DROP COLUMN expensecostcurrency;
ALTER TABLE project DROP COLUMN expenseunitofmeasure;

------------------------------------------------------------------------------------------
-- remove cost from items

-- convert non-billable product items to expenses
-- at this point itemtype values are: service = 0; expense = 1; product = 2
-- converted expenses have the tag '#migration,#non-billable-product'
INSERT INTO expense (expenseid,uuid,created,updated,projectid,date,amount,currency,description,notes,billable,taxexempt,billabletotalamount,billabletotalcurrency,tags) 
SELECT itemid,uuid,created,updated,projectid,date,units*cost,costcurrency,name || ' ' || description,notes,false,false,0,costcurrency,'#migration,#non-billable-product'
FROM item
WHERE item.itemtype=2 AND item.billable=false;

-- delete non-billable product items because they were replaced by expenses
DELETE FROM item WHERE item.itemtype=2 AND item.billable=false;

-- create expenses for billable product items with costs to save their costs
-- converted expenses have the tag '#migration,#product-with-cost'
INSERT INTO expense (expenseid,uuid,created,updated,projectid,date,amount,currency,description,notes,billable,taxexempt,billabletotalamount,billabletotalcurrency,tags) 
SELECT itemid,uuid,created,updated,projectid,date,units*cost,costcurrency,name || ' ' || description,notes,false,false,0,costcurrency,'#migration,#product-with-cost'
FROM item
WHERE item.itemtype=2 AND item.billable=true AND item.cost!=0;

-- create expenses for service items with costs to save their costs
-- converted expenses have the tag '#migration,#service-with-cost'
INSERT INTO expense (expenseid,uuid,created,updated,projectid,date,amount,currency,description,notes,billable,taxexempt,billabletotalamount,billabletotalcurrency,tags) 
SELECT itemid,uuid,created,updated,projectid,date,units*cost,costcurrency,name || ' ' || description,notes || ' Cost: ' || cost || ' ' costcurrency,false,false,0,costcurrency,'#migration,#service-with-cost'
FROM item
WHERE item.itemtype=0 AND item.cost!=0;

-- drop cost columns
ALTER TABLE item DROP COLUMN cost;
ALTER TABLE item DROP COLUMN costcurrency;

ALTER TABLE catalogitem DROP COLUMN cost;
ALTER TABLE catalogitem DROP COLUMN costcurrency;